/*
Under GNU License 2

Read copying in the root directory of the distribution
If you do not have copying, contact me because you should 
have received it with this software package, the author, 
or you can view the license at <http://www.gnu.org>
*/
//import java.lang.*;
import java.net.InetAddress;
import java.net.DatagramSocket;
import java.net.DatagramPacket;
import java.net.UnknownHostException;
import java.net.SocketException;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.StringTokenizer;

/**
MtftpConnection

A class to manage either a read or write tftp request

@author Mark Benvenuto
@version 0.4
@see Mtfp
@see MtftpReadConnection 
@see MtftpWriteConnection 

*/
public class MtftpConnection 
{
	InetAddress cia;
	DatagramSocket conn;
	int localport, remoteport;
	String remotefile, localfile;
	String status;
	int error, debug, maxPacketSize = 516;
	int TransferType;
	static final short TFTP_RRQ = 1;
	static final short TFTP_WRQ = 2;
	static final short TFTP_DATA = 3;
	static final short TFTP_ACK = 4;
	static final short TFTP_ERROR = 5;
	static final String TFTP_netascii = "netascii";
	static final String TFTP_octet = "octet";
	static final String TFTP_mail = "mail";

	/**
		Creates a new instance of MtftpConnection

		@param ia The internet address of the remote host
		@param remoteFilename The filename of the remote file
		@param localFilename The filename of the local file
		@param rtype The request type, 1 for read, 2 for write
		@param dbstate A flag to enable debug/verbose support, 0 for no debuging
		@see java.net.InternetAddress
	*/
	MtftpConnection(InetAddress ia, String remoteFilename, String localFilename, int rtype, int dstate)
	{
		debug = dstate;
		cia = ia;
		remotefile = remoteFilename;
		localfile = localFilename;
		TransferType = rtype;

		//Create Socket
		try{
			conn = new DatagramSocket();
		}catch( SocketException e)
		{
			System.err.println("Datagram Socket Creation Failed: " + e);
			conn = null;
			return;
		}
		
		localport = conn.getLocalPort();

	}
	/**
		Processes the MtftpConnection
	*/
	void processConnection()
	{
		if(conn == null)
			return;
		if(TransferType == 1)  //Read Request - receive file
		{
			//Read the File
			if(debug > 0)
				System.out.println("Reading the file: " + localfile);

			remoteport = 69; //Default TFTP port
	
			if(debug > 0)
				System.out.println("Created Socket: " + conn);

			//Send Read Request
			try{
				conn.send(BuildRequestPacket((short)1, remotefile, TFTP_netascii ));
			}catch( IOException e)
			{
				System.err.println("Error on Sending RRQ Packet: " + e);
			}

			if(debug > 0)
				System.out.println("Sent RRQ Packet");

			//Hande Data
			File outFile = new File(localfile);
			try{
				FileOutputStream fstream = new FileOutputStream(localfile);
				MtftpReadConnection tftpRConn = new MtftpReadConnection(conn, cia, this, fstream, debug);
				Thread th = new Thread(tftpRConn);
				th.start();
			}catch(IOException IOe)
			{
				System.err.println("FileOutputStream Creation failed: " + IOe);
				return;
			}
		}
		else if(TransferType == 2) //Write Request - send the file
		{
			//Send the File
			if(debug > 0)
				System.out.println("Writing the file: " + localfile);

			remoteport = 69; //Default TFTP port
	
			//Hande Data
			//Check here so that we don't initiate a write even though
			//we might not have the file
			File inFile = new File(localfile);
			if(inFile.exists() == false)
			{
				System.out.println("File does not exist");
				return;
			}

			if(debug > 0)
				System.out.println("Created Socket: " + conn);

			//Send Write Request
			try{
				conn.send(BuildRequestPacket((short)2, remotefile, TFTP_netascii ));
			}catch( IOException e)
			{
				System.err.println("Error on Sending WRQ Packet: " + e);
			}

			if(debug > 0)
				System.out.println("Sent WRQ Packet");

			try{
				FileInputStream fstream = new FileInputStream(localfile);
				//BufferedInputStream bfstream = new BufferedInputStream(fstream, 512);
				MtftpWriteConnection tftpWConn = new MtftpWriteConnection(conn, cia, this, fstream, debug);
				Thread th = new Thread(tftpWConn);
				th.start();
			}catch(IOException IOe)
			{
				System.err.println("FileInputStream Creation failed: " + IOe);
				return;
			}
		}
	}
	/**
		Sets the socket timeout for the connection socket 
		which applies also to MtftpReadConnection and MtftpWriteConnection.

		@param time The new socket timeout in milliseconds
		@see MtftpReadConnection 
		@see MtftpWriteConnection 
	*/
	void setSocketTimeout(int time)
	{	
		try{
			conn.setSoTimeout(time);
		}catch( SocketException e)
		{
			System.err.println("Error on setting socket timeout: " + e);
		}
	}	
	/**
		Gets the socket timeout for the connection socket 
		which applies also to MtftpReadConnection and MtftpWriteConnection.

		@return The socket timeout in milliseconds
		@see MtftpReadConnection 
		@see MtftpWriteConnection 
	*/
	int getSocketTimeout() throws SocketException
	{	
		return conn.getSoTimeout();
	}	
	/**
		Creates a proper read/write request packet
		
		@param Opcode Type of request, 1 for read, 2 for write
		@param Filename The filename to pass the other computer
		@param mode The mode to use, usually netascii
		@return Returns a DatagramPacket representing the packet
		@see java.net.DatagramPacket
	*/
	DatagramPacket BuildRequestPacket(short Opcode, String Filename, String Mode)
	{
		byte data[];
		int pos;
		data = new byte[4 + Filename.length() + Mode.length()];
	

		data[0] = 0;
		data[1] = (byte)(0xF & Opcode);
		pos = 2;

		//Store Filename
	        byte bytes[] = Filename.getBytes();
        	System.arraycopy(bytes, 0, data, pos, bytes.length);
	        data[pos + bytes.length] = 0;
		pos += bytes.length + 1;

		//Store Mode
	        bytes = Mode.getBytes();
        	System.arraycopy(bytes, 0, data, pos, bytes.length);
	        data[pos + bytes.length] = 0;
		
		return new DatagramPacket(data, data.length, cia, remoteport);
	}
	/**
		Creates a proper DATA packet
		
		@param block The block number
		@param bytes The data stored in an array of lenght len
		@param len The length of bytes
		@return Returns a DatagramPacket representing the packet
		@see java.net.DatagramPacket
	*/
	DatagramPacket BuildDATAPacket(int block, byte[] bytes, int len)
	{
		byte data[];
		data = new byte[4 + len];

		data[0] = 0;
		data[1] = 3;
		
		data[2] = (byte)(0xFF & (block >> 8));
		data[3] = (byte)(0xFF & block);

		//Store Data
        	System.arraycopy(bytes, 0, data, 4, len);
		
		return new DatagramPacket(data, data.length, cia, remoteport);	
	}
	/**
		Creates a proper ACKnowledge packet
		
		@param block The block number
		@return Returns a DatagramPacket representing the packet
		@see java.net.DatagramPacket
	*/
	DatagramPacket BuildACKPacket(int block)
	{
		byte data[];
		data = new byte[4];

		data[0] = 0;
		data[1] = 4;
		
		data[2] = (byte)(0xFF & (block >> 8));
		data[3] = (byte)(0xFF & block);

		return new DatagramPacket(data, data.length, cia, remoteport);	
	}
	/**
		Creates a proper Error packet
		
		@param ErrorCode A code representing the error
		@param str A string describing the error, could be null
		@return Returns a DatagramPacket representing the packet
		@see java.net.DatagramPacket
	*/
	DatagramPacket BuildErrorPacket(short ErrorCode, String str)
	{
		byte data[];
		int pos;
		data = new byte[5 + str.length()];
	

		data[0] = 0;
		data[1] = 5;

		data[2] = (byte)(0xFF & (ErrorCode >> 8));
		data[3] = (byte)(0xFF & ErrorCode);

		//Store String
	        byte bytes[] = str.getBytes();
        	System.arraycopy(bytes, 0, data, 4, bytes.length);
	        data[4 + bytes.length] = 0;
		
		return new DatagramPacket(data, data.length, cia, remoteport);
	}
	/**
		Send remote computer error packet using <B>BuildErrorPacket</B>
		
		@param ErrorCode A code representing the error
		@param str A string describing the error, could be null
	*/
	void sendError(short ErrorCode, String str)
	{
		try{
			conn.send(BuildErrorPacket(ErrorCode, str));
		}catch( IOException e)
		{
			System.err.println("Error on Sending ERROR Packet: " + e);
		}
	}
	/**
		Checks if packet is error packet
		
		@param data Bytes from a packet to check if it is an error packet. Minimum length 2
		@return True if error packet, false otherwise
	*/
	boolean isError(byte[] data)
	{
		if(data[1] == 5)
			return true;
		return false;
	}
	/**
		Prints error packet. If there is not string in the packet, it uses a standard string defined by the RFC.
		
		@param data Bytes from an error packet
	*/
	void printError(byte[] data)
	{
		System.out.println("Error: " + (short)((data[2] << 8) + data[3]));
		String str = readString(data, 4);
		if(str.length() == 0)
			System.out.println("ErrorStr: " + str);
		else
		{
			switch(data[3])
			{
				case 0:
					System.out.println("Not defined, see error message (if any).");
					break;
				case 1:
					System.out.println("File not found.");
					break;
				case 2:
					System.out.println("Access violation.");
					break;
				case 3:
					System.out.println("Disk full or allocation exceeded.");
					break;
				case 4:
					System.out.println("Illegal TFTP operation.");
					break;
				case 5:
					System.out.println("Unknown transfer ID.");
					break;
				case 6:
					System.out.println("File already exists.");
					break;
				case 7:
					System.out.println("No such user.");
					break;
			}
		}
	}
	/*
		Used to read a string from an arrary of bytes, like a packet
	
		@param rs Bytes from a packet to read for string
		@param off Offset is the position to start reading from
		@return A String from the packet, could be zero length
	*/
	protected String readString(byte[] rs, int off)
	{
		String s = new String();
		off--;
		while(rs[++off]!=(byte)'\0')
			s += String.valueOf((char)rs[off]);
		return s;
	}

}
