/*
Under GNU License 2

Read copying in the root directory of the distribution
If you do not have copying, contact me because you should 
have received it with this software package, the author, 
or you can view the license at <http://www.gnu.org>
*/
//import java.lang.*;
import java.net.InetAddress;
import java.net.DatagramSocket;
import java.net.DatagramPacket;
import java.net.UnknownHostException;
import java.net.SocketException;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.StringTokenizer;
import MtftpConnection;

/**
MtftpReadConnection 

A class to manage an entire read tftp request. 
This class processes the file as a thread.

@author Mark Benvenuto
@version 0.4
@see Mtftp
@see MtftpReadConnection 
@see MtftpWriteConnection 

*/
public class MtftpReadConnection implements Runnable
{
	DatagramSocket s;
	MtftpConnection conn;
	InetAddress ria;
	FileOutputStream ostream;
	int remoteport;
	int maxPacketSize = 516;
	int debug=0;
	int maxACKtries=3;
	/**
		Creates a new instance of MtftpReadConnection

		@param ds A DatagramSocket for i/o
		@param ia The internet address of the remote host
		@param tftpconn A refrence to the calling class
		@param fstream Stream to write file to
		@param dbstate A flag to enable debug/verbose support, 0 for no debuging
		@see java.net.DatagramSocket
		@see java.net.InternetAddress
	*/
	MtftpReadConnection(DatagramSocket ds, InetAddress ia, MtftpConnection tftpconn, FileOutputStream fstream, int dbstate)
	{
			s = ds;
			conn = tftpconn;
			debug = dbstate;
			ria = ia;
			ostream = fstream;
	}
	/**
		Sets maximum retries for sending ACK, default 3

		@param tries Number of retries
	*/
	void setMaxTries(int tries)
	{
		if(tries > 0)
			maxACKtries = tries;
	}
	/**
		Sends a DatagramPacket representing an ACK packet

		@param tries Number of retries
		@return true if success, false otherwise
		@see java.net.DatagramPacket
	*/
	private boolean sendACK(DatagramPacket p)
	{
		try{
			s.send(p);
			return true;
		}catch(IOException e)
		{
			System.err.println("Failed to Send ACK packet: " + e);
		}
		return false;
	}
	/**
		Processes the read connection
	*/
	public void run()
	{
		int isRead = 1, blocknum = 1, oldblocknum = 0, pblocknum = 0;
		byte[] ACKdata = new byte[4];
		int ACKtries = 0, RECVtries = 0;
		boolean ACKretcode = false;
		ACKdata[0] = 0;
		ACKdata[1] = 4;
		byte data[] = new byte[maxPacketSize];
		int datalength = 516;
		DatagramPacket ACKp = new DatagramPacket(ACKdata, ACKdata.length);
		DatagramPacket p = new DatagramPacket(data, maxPacketSize);
		remoteport = 0;
		while(isRead == 1)
		{
			if(debug > 0)
				System.out.println("Receiving Packet");
			try{
				s.receive(p);
			} 
			catch(IOException e)
			{
				System.err.println("Error Receiving Packet: " + e);
				if(RECVtries > maxACKtries)
				{
					System.out.println("Waited to many times for DATA packet");
					return;	
				}
				RECVtries++;
				System.out.println("Continuing to wait for Packet");
				continue;
			}
			if(remoteport == 0)
			{
				remoteport = p.getPort();
				ACKp.setPort(remoteport);
				ACKp.setAddress(ria);
			}
			datalength = p.getLength();
			if(debug > 0)
			{
				System.out.println("Received Packet: " + p);
				System.out.println("Type: " + data[0] + data[1]);
				if(data[1] == 3)
				{
					System.out.println("Received DATA Packet( Len: " + datalength + " )");
					System.out.println(p.getData());
				}
			}
			if(conn.isError(data))
			{
				System.out.println("Received ERROR Packet");
				conn.printError(data);
				isRead = 0;
				break;
			}
			if(data[1] != 3)
			{	
				//Failed to get DATA packet
				conn.sendError((short)4, "Expected DATA Packet");
				System.out.println("Expected DATA Packet");
				break;
			}
			pblocknum = (data[2] << 8) + data[3];
			if(pblocknum == blocknum)
			{
				if(debug > 0)
					System.out.println("Received Correct Block Num: " + pblocknum);
				//conn.sendError((short)0, "Lazy Client");
				//This is the block we were expecting
				oldblocknum++;
				blocknum++;
				ACKtries = 0;

				//Acknowledge the DATA packet
				//We send ACK packet before writing data so that 
				//we can have server get us next DATA packet as we
				//right current one to write data efficiently
				if(debug > 0)
					System.out.println("Acknowleding the Packet");
				ACKdata[2] = data[2];
				ACKdata[3] = data[3];
				ACKretcode = sendACK(ACKp);

				//Write DATA to File
				if(debug > 0)
					System.out.println("Writing DATA to File");
				try{
					ostream.write(data, 4, datalength - 4);
				}catch(IOException IOe)
				{
					System.err.println("Failed to write DATA packet to file: " + IOe);
					return;
				}

				//Exit if sending ACK packet failed here
				//So that the user can have has much of the file as possible
				if(!ACKretcode)
					return;

				if(datalength < 516)
				{
					//Termination Packet
					if(debug > 0)
						System.out.println("Received Termination Packet");
					break;
				}
			}
			else if( pblocknum == oldblocknum )
			{	
				if(debug > 0)
					System.out.println("Received Duplicate Block Num: " + pblocknum);
				//Last ACK failed to receive
				if(ACKtries > maxACKtries)
				{
					//We have retried to many times
					//Send an error message to be nice
					conn.sendError((short)4, "Too Many Retries");

					System.out.println("Failed to Receive proper ACK for DATA packet");

					return;
				}
				//Send it again
				if(debug > 0)
					System.out.println("Acknowleding the Packet");
				if(!sendACK(ACKp))
					return;
				ACKtries++;
			}
			else
			{
				if(debug > 0)
					System.out.println("Received Incorrect Block Num: " + pblocknum);
				//We have an invalid block number
				conn.sendError((short)4, "Invalid BLOCK number");
				return;
			}
		}
	}
}
