/*
Under GNU License 2

Read copying in the root directory of the distribution
If you do not have copying, contact me because you should 
have received it with this software package, the author, 
or you can view the license at <http://www.gnu.org>
*/
//import java.lang.*;
import java.net.InetAddress;
import java.net.DatagramSocket;
import java.net.DatagramPacket;
import java.net.UnknownHostException;
import java.net.SocketException;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.StringTokenizer;
import MtftpConnection;

/**
MtftpWriteConnection 

A class to manage an entire write tftp request. 
This class processes the file as a thread.

@author Mark Benvenuto
@version 0.4
@see Mtftp
@see MtftpReadConnection 
@see MtftpConnection 

*/
public class MtftpWriteConnection implements Runnable
{
	DatagramSocket s;
	MtftpConnection conn;
	InetAddress ria;
	FileInputStream istream;
	int remoteport;
	int maxPacketSize = 516;
	int debug=0;
	int maxDATAtries=3;
	/**
		Creates a new instance of MtftpWriteConnection

		@param ds A DatagramSocket for i/o
		@param ia The internet address of the remote host
		@param tftpconn A refrence to the calling class
		@param fstream Stream to read file from
		@param dbstate A flag to enable debug/verbose support, 0 for no debuging
		@see java.net.DatagramSocket
		@see java.net.InternetAddress
	*/
	MtftpWriteConnection(DatagramSocket ds, InetAddress ia, MtftpConnection tftpconn, FileInputStream fstream, int dbstate)
	{
			s = ds;
			conn = tftpconn;
			debug = dbstate;
			ria = ia;
			istream = fstream;
	}
	/**
		Sets maximum retries for sending DATA, default 3

		@param tries Number of retries
	*/
	void setMaxTries(int tries)
	{
		if(tries > 0)
			maxDATAtries = tries;
	}
	/**
		Sends a DatagramPacket representing an DATA packet

		@param tries Number of retries
		@return true if success, false otherwise
		@see java.net.DatagramPacket
	*/
	private boolean sendDATA(DatagramPacket p)
	{
		try{
			s.send(p);
			return true;
		}catch(IOException e)
		{
			System.err.println("Failed to Send DATA packet: " + e);
		}
		return false;
	}
	/**
		Processes the write connection
	*/
	public void run()
	{
		int isFinished = 0;
		int isRead = 1, blocknum = 1, oldblocknum = 0, pblocknum = 0;
		byte[] ACKdata = new byte[maxPacketSize]; //Also used to receive errors
		int DATAtries = 0, RECVtries = 0;
		byte data[] = new byte[maxPacketSize];
		data[0] = 0;
		data[1] = 3;
		int datalength = 516;
		boolean DATAretcode = false;
		DatagramPacket ACKp = new DatagramPacket(ACKdata, ACKdata.length);
		DatagramPacket p = new DatagramPacket(data, maxPacketSize);
		remoteport = 0;
		while(isRead == 1)
		{
			if(debug > 0)
				System.out.println("Receiving Packet");
			try{
				s.receive(ACKp);
			} 
			catch(IOException e)
			{
				System.err.println("Error Receiving Packet: " + e);
				if(RECVtries > maxDATAtries)
				{
					System.out.println("Waited to many times for ACK packet");
					return;	
				}
				RECVtries++;
				System.out.println("Continuing to wait for Packet");
				continue;
			}
			RECVtries = 0;
			if(remoteport == 0)
			{
				remoteport = ACKp.getPort();
				p.setPort(remoteport);
				p.setAddress(ria);
			}
			datalength = ACKp.getLength();
			if(debug > 0)
			{
				System.out.println("Received Packet: " + ACKp);
				System.out.println("Type: " + ACKdata[0] + ACKdata[1]);
				if(ACKdata[1] == 3)
				{
					System.out.println("Received DATA Packet( Len: " + datalength + " )");
					System.out.println(ACKp.getData());
				}
				if(ACKdata[1] == 4)
				{
					System.out.println("Received ACK Packet( Len: " + datalength + " )");
					System.out.println("ACK: " + (ACKdata[2] << 8) + ACKdata[3]);
				}
			}
			if(conn.isError(ACKdata))
			{
				System.out.println("Received ERROR Packet");
				conn.printError(data);
				isRead = 0;
				break;
			}
			if(ACKdata[1] != 4)
			{	
				//Failed to get ACK packet
				conn.sendError((short)4, "Expected ACK Packet");
				System.out.println("Expected ACK Packet");
				break;
			}
			pblocknum = (ACKdata[2] << 8) + ACKdata[3];
			if(pblocknum == oldblocknum)
			{
				if(isFinished == 1)
					return;
				//Send next DATA packet
				if(debug > 0)
					System.out.println("Received Correct Block Num: " + pblocknum);

				//Read DATA from File
				if(debug > 0)
					System.out.println("Read DATA from File");
				try{
					datalength = istream.read(data, 4, 512);
				}catch(IOException IOe)
				{
					System.err.println("Failed to read DATA packet from file: " + IOe);
					return;
				}
				if(debug > 0)
					System.out.println("Read " + datalength + " bytes");
				if(datalength < 512) //Read the last packet
				{
					isFinished = 1;
					if(datalength < 0)
					{
						datalength = 0;
					}
				}
				p.setLength(datalength + 4);

				DATAtries = 0;
				if(debug > 0)
					System.out.println("Sending the DATA Packet");
				data[2] = (byte)(0xFF & (blocknum >> 8));
				data[3] = (byte)(0xFF & blocknum);
				if(debug > 0)
					System.out.println("Sending Packet #: " + data[2] + data[3]);
				DATAretcode = sendDATA(p);

				//Exit if sending DATA packet failed here
				if(!DATAretcode)
					return;

				//This is the block we were expecting
				oldblocknum++;
				blocknum++;

				continue;
			}
			if( pblocknum < oldblocknum)
			{
				if(debug > 0)
					System.out.println("Received Duplicate Block Num: " + pblocknum);
				//Last DATA failed to receive
				if(DATAtries > maxDATAtries)
				{
					//We have retried to many times
					//Send an error message to be nice
					conn.sendError((short)4, "Too Many Retries");

					System.out.println("Failed to Receive proper ACK for DATA packet");
					if(isFinished == 1)
						System.out.println("File was sent off completely");
					return;
				}
				//Send it again
				if(debug > 0)
					System.out.println("Sending the DATA Packet");
				data[3] = (byte)(data[3] + 1);
				if(!sendDATA(p))
					return;
				DATAtries++;
			}
			else if( pblocknum > oldblocknum)
			{
				if(debug > 0)
					System.out.println("Received Incorrect Block Num: " + pblocknum);
				//We have an invalid block number
				conn.sendError((short)4, "Invalid BLOCK number");
				return;
			}
		}
	}
}